#!/usr/bin/env python

# import sys
# from importlib import reload
# reload(sys)
# sys.setdefaultencoding("ISO-8859-1")

import re
from copy import deepcopy
from bionetgen.atomizer.utils.util import logMess
import string

try:
    # 3.0.0 and above
    from pyparsing import pyparsing_common

    csl = pyparsing_common.comma_separated_list
except ImportError:
    # below 3.0.0
    from pyparsing import commaSeparatedList as csl
import pyparsing
from itertools import dropwhile

# import StringIO
from io import StringIO
import pyparsing as pyp


def evaluatePiecewiseFunction(function):
    pass


def bnglReaction(
    reactant,
    product,
    rate,
    tags,
    translator=[],
    isCompartments=False,
    reversible=True,
    comment="",
    reactionName=None,
):
    finalString = ""
    # if translator != []:
    #    translator = balanceTranslator(reactant,product,translator)
    if len(reactant) == 0 or (len(reactant) == 1 and reactant[0][1] == 0):
        finalString += "0 "
    for index in range(0, len(reactant)):
        tag = ""
        if reactant[index][2] in tags and isCompartments:
            tag = tags[reactant[index][2]]
        translated = printTranslate(reactant[index], tag, translator)
        finalString += translated
        if index < len(reactant) - 1:
            finalString += " + "

    if reversible:
        finalString += " <-> "
    else:
        finalString += " -> "
    if len(product) == 0:
        finalString += "0 "

    for index in range(0, len(product)):
        tag = ""
        if isCompartments:
            if len(product[index]) > 2 and product[index][2] in tags:
                tag = tags[product[index][2]]
        translated = printTranslate(product[index], tag, translator)

        finalString += translated
        if index < len(product) - 1:
            finalString += " + "
    finalString += " " + rate + " " + comment
    finalString = re.sub(r"(\W|^)0\(\)", "0", finalString)
    if reactionName:
        finalString = "{0}: {1}".format(reactionName, finalString)
    return finalString


def printTranslate(chemical, tags, translator={}):
    tmp = []
    if chemical[0] not in translator:
        app = chemical[0] + "()" + tags
    else:
        translator[chemical[0]].addCompartment(tags)
        app = str(translator[chemical[0]])
    if float(int(chemical[1])) == chemical[1]:
        for item in range(0, int(chemical[1])):
            tmp.append(app)
    else:
        idx = logMess(
            "ERROR:SIM205",
            "Cannot deal with non integer stoicheometries: {0}* {1}".format(
                chemical[1], chemical[0]
            ),
        )
        tmp.append(app)
    return " + ".join(tmp)


def balanceTranslator(reactant, product, translator):
    rMolecules = []
    pMolecules = []
    newTranslator = {}
    for species in reactant:
        if species[0] in translator:
            newTranslator[species[0]] = deepcopy(translator[species[0]])
            rMolecules.extend(newTranslator[species[0]].molecules)
    for species in product:
        if species[0] in translator:
            newTranslator[species[0]] = deepcopy(translator[species[0]])
            pMolecules.extend(newTranslator[species[0]].molecules)

    for rMolecule in rMolecules:
        for pMolecule in pMolecules:
            if rMolecule.name == pMolecule.name:
                overFlowingComponents = [
                    x
                    for x in rMolecule.components
                    if x.name not in [y.name for y in pMolecule.components]
                ]
                overFlowingComponents.extend(
                    [
                        x
                        for x in pMolecule.components
                        if x.name not in [y.name for y in rMolecule.components]
                    ]
                )
                rMolecule.removeComponents(overFlowingComponents)
                pMolecule.removeComponents(overFlowingComponents)
    return newTranslator


def rindex(lst, item):
    """
    returns the last ocurrence of an element in alist
    """
    try:
        return next(dropwhile(lambda x: lst[x] != item, reversed(range(len(lst)))))
    except StopIteration:
        raise ValueError("rindex(lst, item): item not in list")


def bnglFunction(
    rule, functionTitle, reactants, compartments=[], parameterDict={}, reactionDict={}
):
    def powParse(match):
        if match.group(1) == "root":
            exponent = "(1/%s)" % match.group(3)
        else:
            exponent = match.group(3)
        if match.group(1) in ["root", "pow"]:
            operator = "^"
        return "({0}){1}({2})".format(match.group(2), operator, exponent)

    def compParse(match):
        translator = {
            "gt": ">",
            "lt": "<",
            "and": "&&",
            "or": "||",
            "geq": ">=",
            "leq": "<=",
            "eq": "==",
            "neq": "!=",
        }
        exponent = match.group(3)
        operator = translator[match.group(1)]
        return "{0} {1} {2}".format(match.group(2), operator, exponent)

    def ceilfloorParse(math):
        flag = False
        if math.group(1) == "ceil":
            flag = True
        if flag:
            return "min(rint({0}+0.5),rint({0} + 1))".format(math.group(2))
        else:
            return "min(rint({0}-0.5),rint({0}+0.5))".format(math.group(2))

    def parameterRewrite(match):
        return match.group(1) + "param_" + match.group(2) + match.group(3)

    def constructFromList(argList, optionList):
        parsedString = ""
        idx = 0
        translator = {
            "gt": ">",
            "lt": "<",
            "and": "&&",
            "or": "||",
            "geq": ">=",
            "leq": "<=",
            "eq": "==",
        }
        while idx < len(argList):
            if type(argList[idx]) is list:
                parsedString += "(" + constructFromList(argList[idx], optionList) + ")"
            elif argList[idx] in optionList:
                if argList[idx] == "ceil":
                    parsedString += "min(rint(({0}) + 0.5),rint(({0}) + 1))".format(
                        constructFromList(argList[idx + 1], optionList)
                    )
                    idx += 1
                elif argList[idx] == "floor":
                    parsedString += "min(rint(({0}) -0.5),rint(({0}) + 0.5))".format(
                        constructFromList(argList[idx + 1], optionList)
                    )
                    idx += 1
                elif argList[idx] in ["pow"]:
                    index = rindex(argList[idx + 1], ",")
                    parsedString += (
                        "(("
                        + constructFromList(argList[idx + 1][0:index], optionList)
                        + ")"
                    )
                    parsedString += (
                        " ^ "
                        + "("
                        + constructFromList(argList[idx + 1][index + 1 :], optionList)
                        + "))"
                    )
                    idx += 1
                elif argList[idx] in ["sqr", "sqrt"]:
                    tag = "1/" if argList[idx] == "sqrt" else ""
                    parsedString += (
                        "(("
                        + constructFromList(argList[idx + 1], optionList)
                        + ") ^ ({0}2))".format(tag)
                    )
                    idx += 1
                elif argList[idx] == "root":
                    index = rindex(argList[idx + 1], ",")
                    tmp = (
                        "1/("
                        + constructFromList(argList[idx + 1][0:index], optionList)
                        + "))"
                    )
                    parsedString += (
                        "(("
                        + constructFromList(argList[idx + 1][index + 1 :], optionList)
                        + ") ^ "
                        + tmp
                    )
                    idx += 1
                elif argList[idx] == "piecewise":
                    index1 = argList[idx + 1].index(",")
                    try:
                        index2 = argList[idx + 1][index1 + 1 :].index(",") + index1 + 1
                        try:
                            index3 = (
                                argList[idx + 1][index2 + 1 :].index(",") + index2 + 1
                            )
                        except ValueError:
                            index3 = -1
                    except ValueError:
                        parsedString += constructFromList(
                            [argList[idx + 1][index1 + 1 :]], optionList
                        )
                        index2 = -1
                    if index2 != -1:
                        condition = constructFromList(
                            [argList[idx + 1][index1 + 1 : index2]], optionList
                        )
                        result = constructFromList(
                            [argList[idx + 1][:index1]], optionList
                        )
                        if index3 == -1:
                            result2 = constructFromList(
                                [argList[idx + 1][index2 + 1 :]], optionList
                            )
                        else:
                            result2 = constructFromList(
                                ["piecewise", argList[idx + 1][index2 + 1 :]],
                                optionList,
                            )
                        parsedString += "if({0},{1},{2})".format(
                            condition, result, result2
                        )
                    idx += 1
                elif argList[idx] in ["and", "or"]:
                    symbolDict = {"and": " && ", "or": " || "}
                    indexArray = [-1]
                    elementArray = []
                    for idx2, element in enumerate(argList[idx + 1]):
                        if element == ",":
                            indexArray.append(idx2)
                    indexArray.append(len(argList[idx + 1]))
                    tmpStr = argList[idx + 1]
                    for idx2, _ in enumerate(indexArray[0:-1]):
                        elementArray.append(
                            constructFromList(
                                tmpStr[indexArray[idx2] + 1 : indexArray[idx2 + 1]],
                                optionList,
                            )
                        )
                    parsedString += symbolDict[argList[idx]].join(elementArray)
                    idx += 1
                elif argList[idx] == "lambda":

                    tmp = "("
                    # ASS2019 - I'm not sure if this is an actual solution or
                    # this should just never happen. argList[idx+1] sometimes
                    # returns _only_ ['0'] and thus the following call fails with
                    # ValueError. Not sure if the list is built wrong or this
                    # result is not handled correctly. Either way, this, for now,
                    # skirts the issue.
                    try:
                        upperLimit = rindex(argList[idx + 1], ",")
                    except ValueError:
                        idx += 1
                        continue
                    parsedParams = []
                    for x in argList[idx + 1][0:upperLimit]:
                        if x == ",":
                            tmp += ", "
                        else:
                            tmp += "param_" + x
                            parsedParams.append(x)

                    # tmp = ''.join([x for x in constructFromList(argList[idx+1][0:upperLimit])])
                    tmp2 = ") = " + constructFromList(
                        argList[idx + 1][rindex(argList[idx + 1], ",") + 1 :],
                        optionList,
                    )
                    for x in parsedParams:
                        while re.search(r"(\W|^)({0})(\W|$)".format(x), tmp2) != None:
                            tmp2 = re.sub(
                                r"(\W|^)({0})(\W|$)".format(x), r"\1param_\2 \3", tmp2
                            )
                    idx += 1
                    parsedString += tmp + tmp2
            else:
                parsedString += argList[idx]
            idx += 1
        return parsedString

    def changeToBNGL(functionList, rule, function):
        oldrule = ""
        # if the rule contains any mathematical function we need to reformat
        while any(
            [
                re.search(r"(\W|^)({0})(\W|$)".format(x), rule) != None
                for x in functionList
            ]
        ) and (oldrule != rule):
            oldrule = rule
            for x in functionList:
                rule = re.sub("({0})\(([^,]+),([^)]+)\)".format(x), function, rule)
            if rule == oldrule:
                logMess("ERROR:TRS001", "Malformed pow or root function %s" % rule)
                print("meep")
        return rule

    # rule = changeToBNGL(['pow','root'],rule,powParse)
    rule = changeToBNGL(["gt", "lt", "leq", "geq", "eq"], rule, compParse)
    # rule = changeToBNGL(['and','or'],rule,compParse)
    flag = True
    contentRule = (
        pyparsing.Word(pyparsing.alphanums + "_")
        | ","
        | "."
        | "+"
        | "-"
        | "*"
        | "/"
        | "^"
        | "&"
        | ">"
        | "<"
        | "="
        | "|"
    )
    parens = pyparsing.nestedExpr("(", ")", content=contentRule)
    finalString = ""
    # remove ceil,floor

    if any(
        [
            re.search(r"(\W|^)({0})(\W|$)".format(x), rule) != None
            for x in ["ceil", "floor", "pow", "sqrt", "sqr", "root", "and", "or"]
        ]
    ):
        argList = parens.parseString("(" + rule + ")").asList()
        rule = constructFromList(
            argList[0], ["floor", "ceil", "pow", "sqrt", "sqr", "root", "and", "or"]
        )

    while "piecewise" in rule:
        argList = parens.parseString("(" + rule + ")").asList()
        rule = constructFromList(argList[0], ["piecewise"])
    # remove references to lambda functions
    if "lambda(" in rule:
        lambdaList = parens.parseString("(" + rule + ")")
        functionBody = constructFromList(lambdaList[0].asList(), ["lambda"])
        flag = False
        rule = "{0}{1}".format(functionTitle, functionBody)

    tmp = rule
    # delete the compartment from the rate function since cBNGL already does it
    # this is not true as seen by the test
    for compartment in compartments:
        # if len(reactants) < 2:
        #    tmp = re.sub('^{0}\s*[*]'.format(compartment[0]),'',tmp)
        #    tmp = re.sub('([*]\s*{0})$'.format(compartment[0]),'',tmp)

        if compartment[0] in tmp:
            tmp = re.sub(
                r"(\W|^)({0})(\W|$)".format(compartment[0]),
                r"\1 {0} \3".format(str(compartment[1])),
                tmp,
            )
            # tmp = re.sub(r'(\W)({0})(\W)'.format(compartment[0]),r'\1%s\3' % str(compartment[1]),tmp)
            # logMess('INFO:MSC005','Exchanging reference to compartment %s for its dimensions' % compartment[0])

    # change references to time for time()
    # tmp =re.sub(r'(\W|^)(time)(\W|$)',r'\1time()\3',tmp)
    # tmp =re.sub(r'(\W|^)(Time)(\W|$)',r'\1time()\3',tmp)
    while re.search(r"(\W|^)inf(\W|$)", tmp) != None:
        tmp = re.sub(r"(\W|^)(inf)(\W|$)", r"\1 1e20 \3", tmp)
    # BNGL has ^ for power.
    if flag:
        finalString = "%s = %s" % (functionTitle, tmp)
    else:
        finalString = tmp
    # change references to local parameters
    for parameter in parameterDict:
        finalString = re.sub(
            r"(\W|^)({0})(\W|$)".format(parameter),
            r"\g<1>{0}\g<3>".format(parameterDict[parameter]),
            finalString,
        )
    # change references to reaction Id's to their netflux equivalent
    for reaction in reactionDict:
        if reaction in finalString:
            finalString = re.sub(
                r"(\W|^)({0})(\W|$)".format(reaction),
                r"\g<1>{0}\g<3>".format(reactionDict[reaction]),
                finalString,
            )

    # combinations '+ -' break ibonetgen
    finalString = re.sub(r"(\W|^)([-])(\s)+", r"\1-", finalString)
    # changing reference of 't' to time()
    # finalString = re.sub(r'(\W|^)(t)(\W|$)',r'\1time()\3',finalString)
    # pi
    finalString = re.sub(r"(\W|^)(pi)(\W|$)", r"\g<1>3.1415926535\g<3>", finalString)
    # print(reactants,finalString)
    # log for log 10
    finalString = re.sub(r"(\W|^)log\(", r"\1 ln(", finalString)
    # reserved keyword: e
    finalString = re.sub(r"(\W|^)(e)(\W|$)", r"\g<1>__e__\g<3>", finalString)
    # changing ceil
    # avoiding variables whose name starts with a number

    # removing mass-action elements

    tmp = finalString

    # print(finalString,reactants)
    # for reactant in reactants:
    #    finalString = re.sub(r'(\W|^)({0}\s+\*)'.format(reactant[0]),r'\1',finalString)
    #    finalString = re.sub(r'(\W|^)(\*\s+{0}(\s|$))'.format(reactant[0]),r'\1',finalString)
    # print(finalString)

    # if finalString != tmp:
    #    logMess('WARNING','Removed mass action elements from )
    return finalString


def curateParameters(param):
    """
    The objective of this function is to remove elements extraneous to bionetgen
    """
    for element in range(0, len(param)):
        tmp = param[element]
        while re.search(r"(\W|^)inf(\W|$)", tmp) != None:
            tmp = re.sub(r"(\W|^)(inf)(\W|$)", r"\g<1>1e20\g<3>", tmp)
        param[element] = tmp
    return param


def finalText(
    comments,
    param,
    molecules,
    species,
    observables,
    rules,
    functions,
    compartments,
    annotations,
    fileName,
):
    # output = open(fileName,'w')
    output = StringIO()
    # output.write(comments.decode('ascii','ignore'))
    output.write(comments)
    output.write("begin model\n")
    param = curateParameters(param)
    output.write(sectionTemplate("parameters", param))
    if len(compartments) > 0:
        output.write(sectionTemplate("compartments", compartments))
    output.write(
        sectionTemplate("molecule types", molecules, annotations["moleculeTypes"])
    )
    output.write(sectionTemplate("seed species", species))
    output.write(sectionTemplate("observables", observables))
    if len(functions) > 0:
        output.write(sectionTemplate("functions", functions))
    output.write(sectionTemplate("reaction rules", rules))
    output.write("end model\n")
    # output.write('generate_network({overwrite=>1})\n')
    # output.write('simulate({method=>"ode",t_end=>100,n_steps=>100})')
    # output.write('writeXML()\n')
    # with open(fileName,'w') as outputFile:
    #    outputFile.write(output.getvalue())
    # output.close()

    return output.getvalue()


def sectionTemplate(name, content, annotations={}):
    section = "begin %s\n" % name
    temp = []
    for line in content:
        if line in annotations:
            for ann in annotations[line]:
                temp.append("\t%s\n" % ann)
        temp.append("\t%s\n" % line)
    # temp = ['\t%s\n' % line for line in content]
    section += "".join(temp)
    section += "end %s\n" % name
    return section


# 341,6,12


def extendFunction(function, subfunctionName, subfunction):
    def constructFromList(argList, optionList, subfunctionParam, subfunctionBody):
        parsedString = ""
        idx = 0
        while idx < len(argList):
            if type(argList[idx]) is list:
                parsedString += (
                    "("
                    + constructFromList(
                        argList[idx], optionList, subfunctionParam, subfunctionBody
                    )
                    + ")"
                )
            elif argList[idx] in optionList:
                tmp = subfunctionBody
                commaIndexes = [0]
                commaIndexes.extend(
                    [i for i, x in enumerate(argList[idx + 1]) if x == ","]
                )
                commaIndexes.append(len(argList[idx + 1]))
                instancedParameters = [
                    argList[idx + 1][commaIndexes[i] : commaIndexes[i + 1]]
                    for i in range(0, len(commaIndexes) - 1)
                ]
                for parameter, instance in zip(subfunctionParam, instancedParameters):
                    if "," in instance:
                        instance.remove(",")
                    parsedParameter = (
                        " ( "
                        + constructFromList(
                            instance, optionList, subfunctionParam, subfunctionBody
                        )
                        + " ) "
                    )
                    tmp = re.sub(
                        r"(\W|^)({0})(\W|$)".format(parameter.strip()),
                        r"\1{0} \3".format(parsedParameter),
                        tmp,
                    )
                parsedString += " " + tmp + " "
                idx += 1
            else:
                if argList[idx] == "=":
                    parsedString += " " + argList[idx] + " "
                else:
                    parsedString += argList[idx]
            idx += 1
        return parsedString

    param = subfunction.split(" = ")[0][len(subfunctionName) + 1 : -1]
    # ASS2019: There are cases where the fuction doesn't have a definition and the
    # following line errors out with IndexError, let's handle it.
    try:
        body = subfunction.split(" = ")[1]
    except IndexError as e:
        logMess(
            "ERROR:TRS002",
            "This function doesn't have a definition, note that atomizer doesn't allow for function linking: {}".format(
                subfunction
            ),
        )
        raise e
    while (
        re.search(r"(\W|^){0}\([^)]*\)(\W|$)".format(subfunctionName), function) != None
    ):
        contentRule = (
            pyparsing.Word(pyparsing.alphanums + "_.")
            | ","
            | "+"
            | "-"
            | "*"
            | "/"
            | "^"
            | "&"
            | ">"
            | "<"
            | "="
            | "|"
        )
        parens = pyparsing.nestedExpr("(", ")", content=contentRule)
        subfunctionList = parens.parseString("(" + function + ")").asList()
        function = constructFromList(
            subfunctionList[0], [subfunctionName], param.split(","), body
        )
    return function



import xmltodict, copy, os, json


class BNGGdiff:
    """
    Class to compare two contact maps generated by

    Usage: BNGGdiff(inp1, inp2).run()
           BNGGdiff(inp1, inp2, out1).run()
           BNGGdiff(inp1, inp2, out1, out2).run()
           BNGGdiff(inp1, inp2, out, mode="matrix").run()

    Arguments
    ---------
    inp1 : str
        path to the first contact map graphml file
    inp2 : str
        path to the second contact map graphml file
    out1 : str
        (optional) path to the output file for inp1 - inp2 graph
    out2 : str
        (optional) path to the second output file for inp2 - inp1 graph
    mode : str
        diff mode, currently available modes are "matrix" and "union"
    """

    def __init__(
        self,
        inp1,
        inp2,
        out=None,
        out2=None,
        mode="matrix",
        colors={
            "g1": ["#dadbfd", "#e6e7fe", "#f3f3ff"],
            "g2": ["#ff9e81", "#ffbfaa", "#ffdfd4"],
            "intersect": ["#c4ed9e", "#d9f4be", "#ecf9df"],
        },
    ) -> None:
        self.input = inp1
        self.input2 = inp2
        self.output = out
        self.output2 = out2
        if isinstance(colors, dict):
            self.colors = colors
        elif isinstance(colors, str):
            # load json
            with open(colors, "r") as f:
                self.colors = json.load(f)
        elif colors is None:
            self.colors = {
                "g1": ["#dadbfd", "#e6e7fe", "#f3f3ff"],
                "g2": ["#ff9e81", "#ffbfaa", "#ffdfd4"],
                "intersect": ["#c4ed9e", "#d9f4be", "#ecf9df"],
            }
        else:
            raise RuntimeError(f"Color type {type(colors)} not recognized")
        self.available_modes = ["matrix", "union"]
        if mode not in self.available_modes:
            raise RuntimeError(
                f"Mode {mode} is not a valid mode, please choose from {self.available_modes}"
            )
        self.mode = mode

        with open(self.input, "r") as f:
            self.gdict_1 = xmltodict.parse(f.read())
        with open(self.input2, "r") as f:
            self.gdict_2 = xmltodict.parse(f.read())

    def diff_graphs(
        self,
        g1,
        g2,
        colors={
            "g1": ["#dadbfd", "#e6e7fe", "#f3f3ff"],
            "g2": ["#c4ed9e", "#d9f4be", "#ecf9df"],
            "intersect": ["#c4ed9e", "#d9f4be", "#ecf9df"],
        },
    ):
        """
        Given two XML dictionaries (using xmltodict) of two graphml
        graphs, do the diff and return the difference graphml xml in
        dictionary format

        The result is g1-g2. By default g1 only stuff are colored green
        g2 only nodes are colored red and common elements are colored blue.
        These can be changed by the colors kwarg which is a dictionary with
        keys g1, g2 and intersect and colors are given as hexcode strings.

        Usage: diff_graphs(g1_dict, g2_dict)
               diff_graphs(g1_dict, g2_dict,
                    colors={"g1": "#hexstr1",
                            "g2": "#hexstr2",
                            "intersect": "#hexstr3"})

        Arguments
        ---------
        g1 : dict
            input dictionary of the input XML file for the first contact map
        g2 : dict
            second input dictionary of the second input XML file.
        colors (opt): dict
            (optional) A dictionary with keys "g1", "g2" and "intersect". The
            values are color hex strings for the colors you want for graph 1,
            graph 2 and the color for common elements between the two graphs.

        Returns
        -------
        diff : dict
            A dictionary of graphs each of which is a dictionary for the XML file
            of the difference graph. Can be converted back to an XML file using
            `xmltodict` function `unparse`. Each key in the dictionary returned by
            this function is the intended file name for that graph.
        """
        # first do a deepcopy so we don't have to
        # manually do add boilerpate
        if self.mode == "matrix":
            iname1 = os.path.basename(self.input).replace(".graphml", "")
            iname2 = os.path.basename(self.input2).replace(".graphml", "")
            if self.output is None:
                self.output = f"{iname1}_{iname2}_diff.graphml"  # set def
            if self.output2 is None:
                self.output2 = f"{iname2}_{iname1}_diff.graphml"  # set def
            graphs = {}
            diff_gml, _ = self._find_diff(g1, g2, colors=colors)
            graphs[self.output] = diff_gml
            # save recolored g1
            g1_recolor_name = os.path.basename(self.input).replace(
                ".graphml", "_recolored.graphml"
            )
            graphs[g1_recolor_name] = self.gdict_1_recolor
            # save recolored g2
            g2_recolor_name = os.path.basename(self.input2).replace(
                ".graphml", "_recolored.graphml"
            )
            graphs[g2_recolor_name] = self.gdict_2_recolor
            # let's do the reverse
            diff_gml_2, _ = self._find_diff(
                g2,
                g1,
                colors={
                    "g1": colors["g2"],
                    "g2": colors["g1"],
                    "intersect": colors["intersect"],
                },
            )
            graphs[self.output2] = diff_gml_2
            return graphs
        elif self.mode == "union":
            graphs = {}
            g1_name = os.path.basename(self.input).replace(".graphml", "")
            # write recolored g2
            g2_name = os.path.basename(self.input2).replace(".graphml", "")
            if self.output is None:
                union_name = f"{g1_name}_{g2_name}_union.graphml"
            else:
                union_name = self.output
            union_gml = self._find_diff_union(g1, g2, colors=colors)
            graphs[union_name] = union_gml
            return graphs
        else:
            raise RuntimeError(
                f"Mode {self.mode} is not a valid mode, please choose from {self.available_modes}"
            )

    def _find_diff_union(
        self,
        g1,
        g2,
        dg=None,
        colors={
            "g1": ["#dadbfd", "#e6e7fe", "#f3f3ff"],
            "g2": ["#c4ed9e", "#d9f4be", "#ecf9df"],
            "intersect": ["#c4ed9e", "#d9f4be", "#ecf9df"],
        },
    ):
        """
        Usage: diff_graphs(g1_dict, g2_dict)
               diff_graphs(g1_dict, g2_dict,
                    colors={"g1": "#hexstr1",
                            "g2": "#hexstr2",
                            "intersect": "#hexstr3"})

        Arguments
        ---------
        g1 : dict
            input dictionary of the input XML file for the first contact map
        g2 : dict
            second input dictionary of the second input XML file.
        dg : dict
            (optional) dictionary to be modified with the difference. If not given it'll
            be a copy of g1 by default.
        colors : dict
            (optional) A dictionary with keys "g1", "g2" and "intersect". The
            values are color hex strings for the colors you want for graph 1,
            graph 2 and the color for common elements between the two graphs.

        Returns
        -------
        diff : dict
            A dictionary for the XML file of the difference graph. Can be converted
            back to an XML file using `xmltodict` function `unparse`.
        """
        # we first want to do the regular diff
        # we'll need to remap g2 names
        dg, rename_map = self._find_diff(g1, g2, dg=dg, colors=colors)

        # now we loop over g2 nodes and add them to dg with the right
        # colors to get the union version
        node_stack = [(["graphml"], [], g2["graphml"])]

        # now we can loop over nodes
        while len(node_stack) > 0:
            dnode = None
            curr_keys, curr_names, curr_node = node_stack.pop(-1)
            # let's take a look at the difference
            dnode = self._get_node_from_names(g1, curr_names)
            if dnode is None and len(curr_names) > 0:
                # this means we don't have this node in diff graph
                # we need to add it in
                dgnode = self._get_node_from_names(dg, curr_names)
                if dgnode is None:
                    curr_dnode = self._add_node_to_graph(
                        curr_node, dg, curr_names, colors=colors, rmap=rename_map
                    )
                else:
                    rename_map[self._get_node_id(curr_node)] = self._get_node_id(dgnode)
            elif dnode is not None and len(curr_names) > 0:
                # we have the same node in g1
                rename_map[self._get_node_id(curr_node)] = self._get_node_id(dnode)
            # if we have graphs in there, add the nodes to the stack
            if "graph" in curr_node.keys():
                # there is a graph in the node, add the nodes to stack
                if isinstance(curr_node["graph"]["node"], list):
                    for inode, node in enumerate(curr_node["graph"]["node"]):
                        ckey = curr_keys + [node["@id"]]
                        node_stack.append(
                            (ckey, curr_names + [self._get_node_name(node)], node)
                        )
                else:
                    ckey = curr_keys + [curr_node["graph"]["node"]["@id"]]
                    node_stack.append(
                        (
                            ckey,
                            curr_names
                            + [self._get_node_name(curr_node["graph"]["node"])],
                            curr_node["graph"]["node"],
                        )
                    )

        # now we add edges, gotta deal with node renaming
        edge_ctr = len(dg["graphml"]["graph"]["edge"])
        for edge in g2["graphml"]["graph"]["edge"]:
            copied_edge = copy.deepcopy(edge)
            copied_edge["@source"] = rename_map[edge["@source"]]
            copied_edge["@target"] = rename_map[edge["@target"]]
            # ensure we don't already have the same edge
            to_add = True
            for dedge in dg["graphml"]["graph"]["edge"]:
                # exact edge?
                if (dedge["@source"] == copied_edge["@source"]) and (
                    dedge["@target"] == copied_edge["@target"]
                ):
                    to_add = False
                    break
                # inverse direction?
                if (dedge["@target"] == copied_edge["@source"]) and (
                    dedge["@source"] == copied_edge["@target"]
                ):
                    to_add = False
                    break
            if to_add:
                copied_edge["@id"] = f"e{edge_ctr}"
                dg["graphml"]["graph"]["edge"].append(copied_edge)
                edge_ctr += 1

        return dg

    def _find_diff(
        self,
        g1,
        g2,
        dg=None,
        colors={
            "g1": ["#dadbfd", "#e6e7fe", "#f3f3ff"],
            "g2": ["#c4ed9e", "#d9f4be", "#ecf9df"],
            "intersect": ["#c4ed9e", "#d9f4be", "#ecf9df"],
        },
    ):
        if dg is None:
            dg = copy.deepcopy(g1)
        # keep track of naming
        rename_map = {}
        # first find differences in nodes
        # FIXME: Check for single nodes before looping
        node_stack = [(["graphml"], [], g1["graphml"])]
        dnode_stack = [(["graphml"], [], dg["graphml"])]
        while len(node_stack) > 0:
            curr_keys, curr_names, curr_node = node_stack.pop(-1)
            curr_dkeys, curr_dnames, curr_dnode = dnode_stack.pop(-1)
            # write down ID map
            rename_map[self._get_node_id(curr_node)] = self._get_node_id(curr_node)
            # let's take a look at the difference
            g2name = None
            g2node = self._get_node_from_names(g2, curr_names)
            if len(curr_names) > 0:
                # let's get IDs and map them
                curr_name = self._get_node_name(curr_node)
                if not (g2node is None):
                    # also check for name
                    if "data" in g2node.keys():
                        g2name = self._get_node_name(g2node)
                        if g2name is not None or curr_name is not None:
                            if g2name == curr_name:
                                # we have the node in g2, we color it appropriately
                                self._color_node(
                                    curr_dnode,
                                    colors["intersect"][self._get_color_id(curr_dnode)],
                                )
                            else:
                                self._color_node(
                                    curr_dnode,
                                    colors["g1"][self._get_color_id(curr_dnode)],
                                )
                else:
                    if "data" in curr_dnode.keys():
                        # we don't have the node in g2, we color it appropriately
                        self._color_node(
                            curr_dnode, colors["g1"][self._get_color_id(curr_dnode)]
                        )
            # if we have graphs in there, add the nodes to the stack
            if "graph" in curr_node.keys():
                # there is a graph in the node, add the nodes to stack
                if isinstance(curr_node["graph"]["node"], list):
                    for inode, node in enumerate(curr_node["graph"]["node"]):
                        ckey = curr_keys + [node["@id"]]
                        node_stack.append(
                            (ckey, curr_names + [self._get_node_name(node)], node)
                        )
                        dnode = curr_dnode["graph"]["node"][inode]
                        dnode_stack.append(
                            (
                                curr_dkeys + [dnode["@id"]],
                                curr_dnames + [self._get_node_name(dnode)],
                                dnode,
                            )
                        )
                else:
                    ckey = curr_keys + [curr_node["graph"]["node"]["@id"]]
                    node_stack.append(
                        (
                            ckey,
                            curr_names
                            + [self._get_node_name(curr_node["graph"]["node"])],
                            curr_node["graph"]["node"],
                        )
                    )
                    dnode_stack.append(
                        (
                            ckey,
                            curr_dnames
                            + [self._get_node_name(curr_dnode["graph"]["node"])],
                            curr_dnode["graph"]["node"],
                        )
                    )
        # let's recolor both graphs
        self.gdict_1_recolor = self._recolor_graph(self.gdict_1, self.colors["g1"])
        self.gdict_2_recolor = self._recolor_graph(self.gdict_2, self.colors["g2"])
        # resize all fonts, this adds +20
        self._resize_fonts(self.gdict_1, 20)
        self._resize_fonts(self.gdict_2, 20)
        self._resize_fonts(dg, 20)
        return dg, rename_map

    def _recolor_graph(self, g, color_list):
        recol_g = copy.deepcopy(g)
        node_stack = [(["graphml"], [], recol_g["graphml"])]
        while len(node_stack) > 0:
            curr_keys, curr_names, curr_node = node_stack.pop(-1)
            if len(curr_names) > 0:
                self._color_node(curr_node, color_list[self._get_color_id(curr_node)])
            # if we have graphs in there, add the nodes to the stack
            if "graph" in curr_node.keys():
                # there is a graph in the node, add the nodes to stack
                if isinstance(curr_node["graph"]["node"], list):
                    for inode, node in enumerate(curr_node["graph"]["node"]):
                        ckey = curr_keys + [node["@id"]]
                        node_stack.append(
                            (ckey, curr_names + [self._get_node_name(node)], node)
                        )
                else:
                    ckey = curr_keys + [curr_node["graph"]["node"]["@id"]]
                    node_stack.append(
                        (
                            ckey,
                            curr_names
                            + [self._get_node_name(curr_node["graph"]["node"])],
                            curr_node["graph"]["node"],
                        )
                    )
        return recol_g

    def _resize_fonts(self, g, add_to_font):
        node_stack = [(["graphml"], [], g["graphml"])]
        while len(node_stack) > 0:
            curr_keys, curr_names, curr_node = node_stack.pop(-1)
            if len(curr_names) > 0:
                self._resize_node_font(curr_node, add_to_font)
            # if we have graphs in there, add the nodes to the stack
            if "graph" in curr_node.keys():
                # there is a graph in the node, add the nodes to stack
                if isinstance(curr_node["graph"]["node"], list):
                    for inode, node in enumerate(curr_node["graph"]["node"]):
                        ckey = curr_keys + [node["@id"]]
                        node_stack.append(
                            (ckey, curr_names + [self._get_node_name(node)], node)
                        )
                else:
                    ckey = curr_keys + [curr_node["graph"]["node"]["@id"]]
                    node_stack.append(
                        (
                            ckey,
                            curr_names
                            + [self._get_node_name(curr_node["graph"]["node"])],
                            curr_node["graph"]["node"],
                        )
                    )

    def _get_node_from_names(self, g, names):
        if "graphml" in g.keys():
            nodes = g["graphml"]["graph"]["node"]
            if len(names) == 0:
                return g["graphml"]
        else:
            nodes = g["graph"]["node"]
            if len(names) == 0:
                return g
        copy_names = copy.copy(names)
        while len(copy_names) > 0:
            found = False
            key = copy_names.pop(0)
            if isinstance(nodes, list):
                for cnode in nodes:
                    cname = self._get_node_name(cnode)
                    if cname == key:
                        found = True
                        node = cnode
                        if "graph" in node.keys():
                            nodes = node["graph"]["node"]
                    if found:
                        break
            else:
                cname = self._get_node_name(nodes)
                if cname == key:
                    found = True
                    node = nodes
                if "graph" in node.keys():
                    nodes = node["graph"]["node"]
        if not found:
            return None
        return node

    def _get_node_properties(self, node):
        if isinstance(node["data"], list):
            found = False
            for datum in node["data"]:
                if "y:ProxyAutoBoundsNode" in datum.keys():
                    gnode = datum["y:ProxyAutoBoundsNode"]["y:Realizers"]["y:GroupNode"]
                    if isinstance(gnode, list):
                        properties = gnode[0]
                    else:
                        properties = gnode
                    found = True
                elif "y:ShapeNode" in datum.keys():
                    snode = datum["y:ShapeNode"]
                    if isinstance(snode, list):
                        properties = snode[0]
                    else:
                        properties = snode
                    found = True
            if not found:
                raise RuntimeError("Can't find properties for nodes")
        else:
            if "y:ProxyAutoBoundsNode" in node["data"].keys():
                properties = node["data"]["y:ProxyAutoBoundsNode"]["y:Realizers"][
                    "y:GroupNode"
                ]
            elif "y:ShapeNode" in node["data"].keys():
                properties = node["data"]["y:ShapeNode"]
            else:
                raise RuntimeError("Can't find properties for nodes")
        return properties

    def _get_node_name(self, node):
        # node['data'] can be a list if there are
        # multiple data types
        properties = self._get_node_properties(node)
        return properties["y:NodeLabel"]["#text"]

    def _get_node_fill(self, node):
        properties = self._get_node_properties(node)
        return properties["y:Fill"]

    def _get_node_color(self, node):
        return self._get_node_fill(node)["@color"]

    def _resize_node_font(self, node, size):
        properties = self._get_node_properties(node)
        properties["y:NodeLabel"]["@fontSize"] = str(size)

    def _get_font_size(self, node):
        properties = self._get_node_properties(node)
        return int(properties["y:NodeLabel"]["@fontSize"])

    def _get_color_id(self, node):
        # FIXME: This should be fixed at bng level by attaching
        # an attribute to graphml node stating the type of node
        # instead of using colors to check the type
        curr_color = self._get_node_color(node)
        if curr_color == "#D2D2D2":
            # grey indicates a species
            cid = 0
        elif curr_color == "#FFFFFF":
            # white indicates a component
            cid = 1
        elif curr_color == "#FFCC00":
            # yellow indicates a state
            cid = 2
        else:
            raise RuntimeError(f"Node color {curr_color} doesn't match known colors")
        return cid

    def _get_node_from_keylist(self, g, keylist):
        copy_keylist = copy.copy(keylist)
        gkey = copy_keylist.pop(0)
        if len(copy_keylist) == 0:
            # we only have "graphml" as key
            return g[gkey]
        # we are out of group nodes
        if "graph" not in g[gkey].keys():
            return None
        # everything up to here is good,
        # loop over to find the node
        nodes = g[gkey]["graph"]["node"]
        while len(copy_keylist) > 0:
            key = copy_keylist.pop(0)
            found = False
            if isinstance(nodes, list):
                for cnode in nodes:
                    if cnode["@id"] == key:
                        found = True
                        node = cnode
                        try:
                            nodes = node["graph"]["node"]
                        except:
                            break
            else:
                if cnode["@id"] == key:
                    found = True
                    node = cnode
            if not found:
                return None
        return node

    def _color_node(self, node, color) -> bool:
        """
        This uses yEd attributes to change the color of a node

        arguments
            node : dict
                the dictionary version of the node XML
            color : dict
                color dictionary with g1/g2/intersect keys and color hex strings as values
        returns
            bool
                True if colored correctly, False if not
        """
        try:
            fill = self._get_node_fill(node)
            fill["@color"] = color
            return True
        except Exception as e:
            print(f"Couldn't color node, error: {e}")
            return False

    def _get_node_text(self, node):
        noded = node["data"]["y:ProxyAutoBoundsNode"]["y:Realizers"]
        for key in noded.keys():
            if "y:" in key:
                return noded[key]["y:NodeLabel"]["#text"]
        return None

    def _get_node_id(self, node):
        if "@id" in node:
            return node["@id"]
        else:
            return None

    def _set_node_id(self, node, idstr) -> bool:
        if "@id" in node:
            node["@id"] = idstr
            return True
        else:
            return False

    def _get_id_list(self, idstr) -> list:
        id_str_list = idstr.split("::")
        id_int_list = [int(x[1:]) for x in id_str_list]
        return id_int_list

    def _get_id_str(self, id_list) -> str:
        return "::".join([f"n{i}" for i in id_list])

    def _add_node_to_graph(self, node, dg, names, colors=None, rmap={}) -> dict:
        node_to_add_to = self._get_node_from_names(dg, names[:-1])
        copied_node = copy.deepcopy(node)
        if colors is not None:
            self._color_node(copied_node, colors["g2"][self._get_color_id(copied_node)])
        if "graph" in node_to_add_to.keys():
            if isinstance(node_to_add_to["graph"]["node"], list):
                # first do renaming
                node_ids = [
                    self._get_node_id(node) for node in node_to_add_to["graph"]["node"]
                ]
                node_lists = [self._get_id_list(idstr) for idstr in node_ids]
                new_id = node_lists[-1]
                new_id[-1] += 1
                new_id = self._get_id_str(new_id)
                self._set_node_id(copied_node, new_id)
                # now we can add
                node_to_add_to["graph"]["node"].append(copied_node)
            else:
                # TODO: check if this is done correctly
                # it's a single node and we need to turn
                # it into a list instead
                copied_original_node = copy.deepcopy(node_to_add_to["graph"]["node"])
                og_node_id = self._get_node_id(copied_original_node)
                new_id = self._get_id_list(og_node_id)
                new_id[-1] += 1
                new_id = self._get_id_str(new_id)
                self._set_node_id(copied_node, new_id)
                nodes_to_add = [copied_original_node, copied_node]
                node_to_add_to["graph"]["node"] = nodes_to_add
            # add to rename map
            rmap[self._get_node_id(node)] = self._get_node_id(copied_node)
            # TODO: Need to get in there and rename and recolor each
            # node under the one we just copied
            if "graph" in copied_node:
                # let's rename the graph
                if "@id" in copied_node["graph"]:
                    copied_node["graph"]["@id"] = self._get_node_id(copied_node) + ":"
                node_stack = [([], [], copied_node)]
                while len(node_stack) > 0:
                    curr_keys, curr_names, curr_node = node_stack.pop(-1)
                    # Do stuff here
                    # we need to recolor, re-ID each node and add to rename map
                    if len(curr_names) > 0:
                        parent_node = self._get_node_from_names(
                            copied_node, curr_names[:-1]
                        )
                        if colors is not None:
                            self._color_node(
                                curr_node, colors["g2"][self._get_color_id(curr_node)]
                            )
                        parent_node_id = self._get_node_id(parent_node)
                        new_id = self._get_id_list(parent_node_id)
                        curr_id = self._get_id_list(self._get_node_id(curr_node))
                        new_id += [curr_id[-1]]
                        new_id = self._get_id_str(new_id)
                        self._set_node_id(curr_node, new_id)
                        rmap[self._get_id_str(curr_id)] = new_id
                    # if we have graphs in there, add the nodes to the stack
                    if "graph" in curr_node.keys():
                        # there is a graph in the node, add the nodes to stack
                        if isinstance(curr_node["graph"]["node"], list):
                            for inode, node in enumerate(curr_node["graph"]["node"]):
                                ckey = curr_keys + [node["@id"]]
                                node_stack.append(
                                    (
                                        ckey,
                                        curr_names + [self._get_node_name(node)],
                                        node,
                                    )
                                )
                        else:
                            ckey = curr_keys + [curr_node["graph"]["node"]["@id"]]
                            node_stack.append(
                                (
                                    ckey,
                                    curr_names
                                    + [self._get_node_name(curr_node["graph"]["node"])],
                                    curr_node["graph"]["node"],
                                )
                            )
        return copied_node

    def run(self) -> dict:
        # Now we have the graphml files, now we do diff
        graphs = self.diff_graphs(self.gdict_1, self.gdict_2, self.colors)
        for graph_name in graphs.keys():
            # now write gml as graphml
            with open(graph_name, "w") as f:
                xmltodict.unparse(graphs[graph_name], output=f, pretty=True)
        return graphs
